@@ -33,7 +33,7 @@
  *  purpose: read config options from file and store memory
  */
 void
-read_conf(int prog)
+read_conf(unsigned int prog)
 {
   
   memset(confbuf, 0x00, 256);
@@ -71,8 +71,8 @@ read_conf(int prog)
       if((CHROOT=malloc(strlen(extract_array[0])+1)) != NULL)
         strcpy(CHROOT, extract_array[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
 
@@ -111,8 +111,20 @@ read_conf(int prog)
       if((BINDHOST=malloc(strlen(extract_array[0])+1)) != NULL)
         strcpy(BINDHOST,extract_array[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
+      }
+    }
+    
+    /* CONNECTION ACL */
+    if(strncmp(confbuf, "CONN_ACL=", 9) == 0)
+    { 
+      extract(0, confbuf, 8);
+      if((CONN_ACL=malloc(strlen(extract_array[0])+1)) != NULL)
+        strcpy(CONN_ACL,extract_array[0]);
+      else {
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
@@ -123,8 +135,8 @@ read_conf(int prog)
       if((PIDFILE=malloc(strlen(extract_array[0])+1)) != NULL)
         strcpy(PIDFILE,extract_array[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
@@ -135,13 +147,27 @@ read_conf(int prog)
       GREYLISTING=atol(extract_array[0]);
     }
 
+    /* GREYLIST_HOSTADDR */
+    if(strncmp(confbuf, "GREYLIST_HOSTADDR=", 18) == 0)
+    {
+      extract(0, confbuf, 17);
+      GREYLIST_HOSTADDR=atol(extract_array[0]);
+    }
+
     /* TRAINING_MODE */
-    if(strncmp(confbuf, "TRAINING_MODE=", 13) == 0)
+    if(strncmp(confbuf, "TRAINING_MODE=", 14) == 0)
     { 
-      extract(0, confbuf, 12);
+      extract(0, confbuf, 13);
       TRAINING_MODE=atol(extract_array[0]);
     }
 
+    /* TRAINING_POLICY_TIMEOUT */
+    if(strncmp(confbuf, "TRAINING_POLICY_TIMEOUT=", 24) == 0)
+    {
+      extract(0, confbuf, 23);
+      TRAINING_POLICY_TIMEOUT=extract_seconds(extract_array[0]);
+    }
+
     /* GREYLIST_X_HEADER */
     if(strncmp(confbuf, "GREYLIST_X_HEADER=", 18) == 0)
     { 
@@ -205,6 +231,34 @@ read_conf(int prog)
       SPAMTRAP_AUTO_EXPIRE=extract_seconds(extract_array[0]);
     }
     
+    /* HELO CHECK */
+    if(strncmp(confbuf, "HELO_CHECK=", 11) == 0)
+    { 
+      extract(0, confbuf, 10);
+      HELO_CHECK=atol(extract_array[0]);
+    }
+    
+    /* HELO_MAX_COUNT */
+    if(strncmp(confbuf, "HELO_MAX_COUNT=", 15) == 0)
+    { 
+      extract(0, confbuf, 14);
+      HELO_MAX_COUNT=atol(extract_array[0]);
+    }
+    
+    /* HELO_BLACKLIST_AUTO_EXPIRE */
+    if(strncmp(confbuf, "HELO_BLACKLIST_AUTO_EXPIRE=", 27) == 0)
+    { 
+      extract(0, confbuf, 26);
+      HELO_BLACKLIST_AUTO_EXPIRE=extract_seconds(extract_array[0]);
+    }
+    
+    /* HELO_AUTO_EXPIRE */
+    if(strncmp(confbuf, "HELO_AUTO_EXPIRE=", 17) == 0)
+    { 
+      extract(0, confbuf, 16);
+      HELO_AUTO_EXPIRE=extract_seconds(extract_array[0]);
+    }
+    
     /* BLACKLISTING */
     if(strncmp(confbuf, "BLACKLISTING=", 13) == 0)
     { 
@@ -212,6 +266,13 @@ read_conf(int prog)
       BLACKLISTING=atol(extract_array[0]);
     }
 
+    /* AUTO_BLACK_LISTING */
+    if(strncmp(confbuf, "AUTO_BLACK_LISTING=", 19) == 0)
+    { 
+      extract(0, confbuf, 18);
+      AUTO_BLACK_LISTING=atol(extract_array[0]);
+    }
+    
     /* BLACKLIST_HELO */
     if(strncmp(confbuf, "BLACKLIST_HELO=", 15) == 0)
     { 
@@ -254,13 +315,41 @@ read_conf(int prog)
       BLACKLIST_NETBLOCK=atol(extract_array[0]);
     }
     
+    /* BLACKLIST SENDER */
+    if(strncmp(confbuf, "BLACKLISTSENDER=", 16) == 0)
+    { 
+      extract(0, confbuf, 15);
+      BLACKLISTSENDER=atol(extract_array[0]);
+    }
+
+    /* BLACKLIST DNS NAME */
+    if(strncmp(confbuf, "BLACKLISTDNSNAME=", 17) == 0)
+    {
+      extract(0, confbuf, 16);
+      BLACKLISTDNSNAME=atol(extract_array[0]);
+    }
+     
     /* WHITELIST NULL SENDER */
     if(strncmp(confbuf, "WHITELISTNULL=", 14) == 0)
     { 
-      extract(0, confbuf, 12);
+      extract(0, confbuf, 13);
       WHITELISTNULL=atol(extract_array[0]);
     }
 
+    /* WHITELIST SENDER */
+    if(strncmp(confbuf, "WHITELISTSENDER=", 16) == 0)
+    { 
+      extract(0, confbuf, 15);
+      WHITELISTSENDER=atol(extract_array[0]);
+    }
+
+    /* WHITELIST DNS NAME */
+    if(strncmp(confbuf, "WHITELISTDNSNAME=", 17) == 0)
+    { 
+      extract(0, confbuf, 16);
+      WHITELISTDNSNAME=atol(extract_array[0]);
+    }
+
     /* SENDERTHROTTLE */
     if(strncmp(confbuf, "SENDERTHROTTLE=", 15) == 0)
     { 
@@ -282,14 +371,21 @@ read_conf(int prog)
       SENDER_THROTTLE_HOST=atol(extract_array[0]);
     }
     
-    /* MAXIMUM MESSAGES LIMIT */
+    /* MAXIMUM SENDER MESSAGE LIMIT */
     if(strncmp(confbuf, "SENDERMSGLIMIT=", 15) == 0)
     { 
       extract(0, confbuf, 14);
       SENDERMSGLIMIT=atol(extract_array[0]);
     }
     
-    /* MAX VOLUME/QUOTA SENDER TIME LIMIT */
+    /* MAXIMUM SENDER RCPT LIMIT */
+    if(strncmp(confbuf, "SENDERRCPTLIMIT=", 16) == 0)
+    { 
+      extract(0, confbuf, 15);
+      SENDERRCPTLIMIT=atol(extract_array[0]);
+    }
+    
+    /* MAX SENDER VOLUME/QUOTA SENDER TIME LIMIT */
     if(strncmp(confbuf, "SENDERQUOTALIMIT=", 17) == 0)
     { 
       extract(0, confbuf, 16);
@@ -303,12 +399,47 @@ read_conf(int prog)
       SENDERMSGSIZE=atol(extract_array[0]);
     }
     
-    /* MAX TIME LIMIT */
+    /* MAX SENDER TIME LIMIT */
     if(strncmp(confbuf, "SENDERTIMELIMIT=", 16) == 0)
     { 
       extract(0, confbuf, 15);
       SENDERTIMELIMIT=extract_seconds(extract_array[0]);
     }
+
+    /* INACTIVE SENDER EXPIRATION TIME LIMIT */
+    if(strncmp(confbuf, "SENDER_INACTIVE_EXPIRE=", 23) == 0)
+    {
+      extract(0, confbuf, 22);
+      SENDER_INACTIVE_EXPIRE=extract_seconds(extract_array[0]);
+    }
+
+    /* RECIPIENTTHROTTLE */
+    if(strncmp(confbuf, "RECIPIENTTHROTTLE=", 18) == 0)
+    { 
+      extract(0, confbuf, 17);
+      RECIPIENTTHROTTLE=atol(extract_array[0]);
+    }
+    
+    /* MAX RECIPIENT MESSAGES LIMIT */
+    if(strncmp(confbuf, "RECIPIENTMSGLIMIT=", 18) == 0)
+    { 
+      extract(0, confbuf, 17);
+      RECIPIENTMSGLIMIT=atol(extract_array[0]);
+    }
+
+    /* MAX RECIPIENT TIME LIMIT */
+    if(strncmp(confbuf, "RECIPIENTTIMELIMIT=", 19) == 0)
+    { 
+      extract(0, confbuf, 18);
+      RECIPIENTTIMELIMIT=extract_seconds(extract_array[0]);
+    }
+    
+    /* INACTIVE RECIPIENT EXPIRATION TIME LIMIT */
+    if(strncmp(confbuf, "RECIPIENT_INACTIVE_EXPIRE=", 26) == 0)
+    {
+      extract(0, confbuf, 25);
+      RECIPIENT_INACTIVE_EXPIRE=extract_seconds(extract_array[0]);
+    }
     
     /* TRIPLET TIMEOUT */
     if(strncmp(confbuf, "TRIPLET_TIME=", 13) == 0)
@@ -362,48 +493,60 @@ read_conf(int prog)
     /* MYSQL HOST */
     if(strncmp(confbuf, "MYSQLHOST=", 10) == 0)
     { 
-      extract(0, confbuf, 9);
-      if((MYSQLHOST=malloc(strlen(extract_array[0])+1)) != NULL)
-        strcpy(MYSQLHOST, extract_array[0]);
+      extract_conf(0, confbuf, 9);
+      if((MYSQLHOST=malloc(strlen(extract_array_conf[0])+1)) != NULL)
+        strcpy(MYSQLHOST, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
 
     /* MYSQL DATABASE */
     if(strncmp(confbuf, "MYSQLDBASE=", 11) == 0)
     { 
-      extract(0, confbuf, 10);
-      if((MYSQLDBASE=malloc(strlen(extract_array[0])+1)) != NULL)
-        strcpy(MYSQLDBASE, extract_array[0]);
+      extract_conf(0, confbuf, 10);
+      if((MYSQLDBASE=malloc(strlen(extract_array_conf[0])+1)) != NULL)
+        strcpy(MYSQLDBASE, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
 
     /* MYSQL USER */
     if(strncmp(confbuf, "MYSQLUSER=", 10) == 0)
     { 
-      extract(0, confbuf, 9);
-      if((MYSQLUSER=malloc(strlen(extract_array[0])+1)) != NULL)
-        strcpy(MYSQLUSER, extract_array[0]);
+      extract_conf(0, confbuf, 9);
+      if((MYSQLUSER=malloc(strlen(extract_array_conf[0])+1)) != NULL)
+        strcpy(MYSQLUSER, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
 
     /* MYSQL PASS */
     if(strncmp(confbuf, "MYSQLPASS=", 10) == 0)
     { 
-      extract(0, confbuf, 9);
-      if((MYSQLPASS=malloc(strlen(extract_array[0])+1)) != NULL)
-        strcpy(MYSQLPASS, extract_array[0]);
+      extract_conf(0, confbuf, 9);
+      if((MYSQLPASS=malloc(strlen(extract_array_conf[0])+1)) != NULL)
+        strcpy(MYSQLPASS, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
+      }
+    }
+    
+    /* MYSQL OPTIONS */
+    if(strncmp(confbuf, "MYSQLOPT=", 9) == 0)
+    { 
+      extract_conf(0, confbuf, 8);
+      if((MYSQLOPT=malloc(strlen(extract_array_conf[0])+1)) != NULL)
+        strcpy(MYSQLOPT, extract_array_conf[0]);
+      else {
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
 
@@ -417,68 +560,96 @@ read_conf(int prog)
     /* GREYLIST_REJECTION */
     if(strncmp(confbuf, "GREYLIST_REJECTION=", 19) == 0)
     { 
-      extract(0, confbuf, 18);
+      extract_conf(0, confbuf, 18);
       if((postfix_greylist=malloc(512)) != NULL)
-        snprintf(postfix_greylist, 512, "%s %s\n\n", POSTFIX_GREYLIST, extract_array[0]);
+        snprintf(postfix_greylist, 512, "%s %s\n\n", POSTFIX_GREYLIST, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
     /* BLACKLIST_REJECTION */
     if(strncmp(confbuf, "BLACKLIST_REJECTION=", 20) == 0)
     { 
-      extract(0, confbuf, 19);
+      extract_conf(0, confbuf, 19);
       if((postfix_blacklist=malloc(512)) != NULL)
       {
-	if(BLACKLIST_TEMP_REJECT==1)
-          snprintf(postfix_blacklist, 512, "%s %s\n\n", POSTFIX_BLACKLIST_TEMP, extract_array[0]);
-	else
-          snprintf(postfix_blacklist, 512, "%s %s\n\n", POSTFIX_BLACKLIST_PERM, extract_array[0]);
+        if(BLACKLIST_TEMP_REJECT==1)
+          snprintf(postfix_blacklist, 512, "%s %s\n\n", POSTFIX_BLACKLIST_TEMP, extract_array_conf[0]);
+        else
+          snprintf(postfix_blacklist, 512, "%s %s\n\n", POSTFIX_BLACKLIST_PERM, extract_array_conf[0]);
       }
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
     /* THROTTLE BAD SIZE */
-    if(strncmp(confbuf, "SENDER_SIZE_REJECTION=", 16) == 0)
+    if(strncmp(confbuf, "SENDER_SIZE_REJECTION=", 22) == 0)
     { 
-      extract(0, confbuf, 15);
+      extract_conf(0, confbuf, 21);
       if((postfix_bad_size=malloc(512)) != NULL)
-        snprintf(postfix_bad_size, 512, "%s %s\n\n", POSTFIX_BAD_SIZE, extract_array[0]);
+        snprintf(postfix_bad_size, 512, "%s %s\n\n", POSTFIX_BAD_SIZE, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
     /* SPAMTRAP */
-    if(strncmp(confbuf, "SPAMTRAP_REJECTION=", 16) == 0)
+    if(strncmp(confbuf, "SPAMTRAP_REJECTION=", 19) == 0)
     { 
-      extract(0, confbuf, 15);
+      extract_conf(0, confbuf, 18);
       if((postfix_spamtrap=malloc(512)) != NULL)
-        snprintf(postfix_spamtrap, 512, "%s %s\n\n", POSTFIX_SPAMTRAP, extract_array[0]);
+        snprintf(postfix_spamtrap, 512, "%s %s\n\n", POSTFIX_SPAMTRAP, extract_array_conf[0]);
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
-    /* MAX QUOTA EXCEEDED */
-    if(strncmp(confbuf, "SENDER_QUOTA_REJECTION=", 16) == 0)
+    /* MAX SENDER QUOTA EXCEEDED */
+    if(strncmp(confbuf, "SENDER_QUOTA_REJECTION=", 23) == 0)
     { 
-      extract(0, confbuf, 15);
-      if((postfix_quota_exceeded=malloc(512)) != NULL)
-        snprintf(postfix_quota_exceeded, 512, "%s %s\n\n", POSTFIX_QUOTA_EXCEEDED, extract_array[0]);
+      extract_conf(0, confbuf, 22);
+      if((postfix_sender_quota_exceeded=malloc(512)) != NULL)
+      {
+        if(QUOTA_EXCEEDED_TEMP_REJECT==1)
+          snprintf(postfix_sender_quota_exceeded, 512, "%s %s\n\n", POSTFIX_QUOTA_EXCEEDED_TEMP, extract_array_conf[0]);
+        else
+          snprintf(postfix_sender_quota_exceeded, 512, "%s %s\n\n", POSTFIX_QUOTA_EXCEEDED_PERM, extract_array_conf[0]);
+      }
       else {
-	logmessage("malloc(): %s\n", strerror(errno)); 
-	exit(-1);
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
       }
     }
     
+    /* MAX RECIPIENT QUOTA EXCEEDED */
+    if(strncmp(confbuf, "RECIPIENT_QUOTA_REJECTION=", 26) == 0)    { 
+      extract_conf(0, confbuf, 25);
+      if((postfix_recipient_quota_exceeded=malloc(512)) != NULL)
+      {
+        if(QUOTA_EXCEEDED_TEMP_REJECT==1)
+          snprintf(postfix_recipient_quota_exceeded, 512, "%s %s\n\n", POSTFIX_QUOTA_EXCEEDED_TEMP, extract_array_conf[0]);
+        else
+          snprintf(postfix_recipient_quota_exceeded, 512, "%s %s\n\n", POSTFIX_QUOTA_EXCEEDED_PERM, extract_array_conf[0]);
+      }
+      else {
+        logmessage("malloc(): %s\n", strerror(errno)); 
+        exit(-1);
+      }
+    }
+    
+    /* QUOTA_EXCEEDED_TEMP_REJECT */
+    if(strncmp(confbuf, "QUOTA_EXCEEDED_TEMP_REJECT=", 27) == 0)
+    { 
+      extract(0, confbuf, 26);
+      QUOTA_EXCEEDED_TEMP_REJECT=atol(extract_array[0]);
+    }
+
     memset(confbuf, 0x00, 256);
   }
 
@@ -496,11 +667,11 @@ read_conf(int prog)
   /* background policyd */
   if(DAEMON)
   {
-	
+
     /* dont let cleanup run in the background */
     if(prog == 0)
     {
-      if(daemonize(1,1) == -1)
+      if(daemonize(0,0) == -1)
       {
         fprintf(stderr, "daemon(): %s\n", strerror(errno));
         exit(-1);
@@ -509,61 +680,104 @@ read_conf(int prog)
   }
   
   /* dump all debugging info */
-  if(DEBUG)
+  if(DEBUG > 0)
   {
     logmessage(" ---- DAEMON CONFIG ----\n"); 
     logmessage("config: version> %s\n", VERSION);
     logmessage("config: debug> %d\n", DEBUG);
     logmessage("config: daemon mode> %d\n", DAEMON);
+    logmessage("config: bindhost> %s\n", BINDHOST);
+    logmessage("config: bindport> %d\n", BINDPORT);
+    logmessage("config: pidfile> %s\n", PIDFILE);
+    logmessage("config: syslog> %d\n", SYSLOG_FACILITY);
     logmessage("config: chroot> %s\n", CHROOT);
     logmessage("config: uid> %d\n", UID);
     logmessage("config: gid> %d\n", GID);
-    logmessage("config: bindport> %d\n", BINDPORT);
-    logmessage("config: bindhost> %s\n", BINDHOST);
-    logmessage("config: pidfile> %s\n", PIDFILE);
-    logmessage("config: syslog> %d\n\n", SYSLOG_FACILITY);
+    logmessage("config: conn acl> %s\n", CONN_ACL);
+    logmessage("\n");
     
-    logmessage(" ---- MISC CONFIG ----\n"); 
+    logmessage(" ---- DATABASE CONFIG ----\n");
+    logmessage("config: host> %s\n", MYSQLHOST);
+    logmessage("config: user> %s\n", MYSQLUSER);
+    logmessage("config: pass> %s\n", MYSQLPASS);
+    logmessage("config: database> %s\n", MYSQLDBASE);
+    logmessage("config: options> %s\n", MYSQLOPT);
     logmessage("config: failsafe> %d\n", FAILSAFE);
+    logmessage("config: keep alive> %d\n", DATABASE_KEEPALIVE);
+    logmessage("config: version> %d\n", MYSQL_VERSION_ID);
+    logmessage("\n");
+
+    logmessage(" ---- WHITELISTING ----\n"); 
     logmessage("config: whitelisting> %d\n", WHITELISTING);
-    logmessage("config: whitenullsender> %d\n", WHITELISTNULL);
-    logmessage("config: autowhitelist_expire> %d\n", AUTO_WHITELIST_EXPIRE);
-    logmessage("config: autoblacklist_expire> %d\n", AUTO_BLACKLIST_EXPIRE);
+    logmessage("config: whitelistnullsender> %d\n", WHITELISTNULL);
+    logmessage("config: whitelistsender> %d\n", WHITELISTSENDER);
+    logmessage("config: whitelistdnsname> %d\n", WHITELISTDNSNAME);
+    logmessage("config: autowhitelisting> %d\n", AUTO_WHITE_LISTING);
     logmessage("config: autowhitelist_number> %d\n", AUTO_WHITELIST_NUMBER);
+    logmessage("config: autowhitelist_netblock> %d\n", AUTO_WHITELIST_NETBLOCK);
+    logmessage("config: autowhitelist_expire> %d\n", AUTO_WHITELIST_EXPIRE);
+    logmessage("\n");
+
+    logmessage(" ---- BLACKLISTING ----\n"); 
     logmessage("config: blacklisting> %d\n", BLACKLISTING);
     logmessage("config: blacklisting_temp_reject> %d\n", BLACKLIST_TEMP_REJECT);
     logmessage("config: blacklisting_netblock> %d\n", BLACKLIST_NETBLOCK);
-    logmessage("config: autowhitelist_netblock> %d\n", AUTO_WHITELIST_NETBLOCK);
+    logmessage("config: postfix_blacklist> %s\n", postfix_blacklist);
+    logmessage("config: autoblacklisting> %d\n", AUTO_BLACK_LISTING);
     logmessage("config: autoblacklist_number> %d\n", AUTO_BLACKLIST_NUMBER);
-    logmessage("config: postfix_blacklist> %s\n\n", postfix_blacklist);
+    logmessage("config: autoblacklist_expire> %d\n", AUTO_BLACKLIST_EXPIRE);
+    logmessage("config: blacklist_rejection> %s\n", postfix_blacklist);
+    logmessage("\n");
     
-    logmessage(" ---- GREYLISTING CONFIG ----\n"); 
-    logmessage("config: greylisting> %d\n", GREYLISTING);
-    logmessage("config: trainingmode> %d\n", TRAINING_MODE);
-    logmessage("config: greylist_x_header> %d\n", GREYLIST_X_HEADER);
-    logmessage("config: triplet timeout> %d\n", TRIPLET_TIME);
-    logmessage("config: optin/optout> %d\n", OPTINOUT);
-    logmessage("config: postfix_greylist> %s\n\n", postfix_greylist);
-
-    logmessage(" ---- THROTTLE CONFIG ----\n"); 
-    logmessage("config: sender throttle> %d\n", SENDERTHROTTLE);
-    logmessage("config: sender msglimit> %d\n", SENDERMSGLIMIT);
-    logmessage("config: sender quotalimit> %d\n", SENDERQUOTALIMIT);
-    logmessage("config: sender msgsize> %d\n", SENDERMSGSIZE);
-    logmessage("config: sender timelimit> %d\n", SENDERTIMELIMIT);
-    logmessage("config: postfix_bad_size> %s\n", postfix_bad_size);
-    logmessage("config: sender throttle sasl> %d\n", SENDER_THROTTLE_SASL);
-    logmessage("config: sender throttle host> %d\n", SENDER_THROTTLE_HOST);
-    logmessage("config: postfix_quota_exceeded> %s\n\n", postfix_quota_exceeded);
+    logmessage(" ---- HELO (HRP) ----\n");
+    logmessage("config: helo> %d\n", HELO_CHECK);
+    logmessage("config: helo_max_count> %d\n", HELO_MAX_COUNT);
+    logmessage("config: helo_blacklist_auto_expire> %d\n", HELO_BLACKLIST_AUTO_EXPIRE);
+    logmessage("config: helo_auto_expire> %d\n", HELO_AUTO_EXPIRE);
+    logmessage("\n");
 
     logmessage(" ---- SPAMTRAP CONFIG ----\n"); 
     logmessage("config: spamtrap> %d\n", SPAMTRAPPING);
+    logmessage("config: postfix_spamtrap> %s\n", postfix_spamtrap);
     logmessage("config: spamtrapauto_expire> %d\n", SPAMTRAP_AUTO_EXPIRE);
-    logmessage("config: postfix_spamtrap> %s\n\n", postfix_spamtrap);
+    logmessage("\n");
 
-    logmessage(" ---- BLACKLIST HELO CONFIG ----\n"); 
-    logmessage("config: blacklist helo> %d\n", BLACKLIST_HELO);
-    logmessage("config: blacklist helo expire> %d\n", BLACKLIST_HELO_AUTO_EXPIRE);
+    logmessage(" ---- GREYLISTING CONFIG ----\n"); 
+    logmessage("config: greylisting> %d\n", GREYLISTING);
+    logmessage("config: greylist_hostaddr> %d\n", GREYLIST_HOSTADDR);
+    logmessage("config: postfix_greylist> %s\n", postfix_greylist);
+    logmessage("config: greylist_x_header> %d\n", GREYLIST_X_HEADER);
+    logmessage("config: trainingmode> %d\n", TRAINING_MODE);
+    logmessage("config: training_policyd_timeout> %d\n", TRAINING_POLICY_TIMEOUT);
+    logmessage("config: triplet timeout> %d\n", TRIPLET_TIME);
+    logmessage("config: optin/optout> %d\n", OPTINOUT);
+    logmessage("config: optin all in> %d\n", OPTINOUTALL);
+    logmessage("config: triplet auth timeout> %d\n", TRIPLET_AUTH_TIMEOUT);
+    logmessage("config: triplet unauth timeout> %d\n", TRIPLET_UNAUTH_TIMEOUT);
+    logmessage("\n");
+
+    logmessage(" ---- SENDER THROTTLE CONFIG ----\n"); 
+    logmessage("config: sender throttle> %d\n", SENDERTHROTTLE);
+    logmessage("config: sender throttle sasl> %d\n", SENDER_THROTTLE_SASL);
+    logmessage("config: sender throttle host> %d\n", SENDER_THROTTLE_HOST);
+    logmessage("config: postfix_sender_quota_exceeded> %s\n", postfix_sender_quota_exceeded);
+    logmessage("config: quota_exceeded_temp_reject> %d\n", QUOTA_EXCEEDED_TEMP_REJECT);
+    logmessage("config: postfix_bad_size> %s\n", postfix_bad_size);
+    logmessage("config: sender msglimit> %d\n", SENDERMSGLIMIT);
+    logmessage("config: sender quotalimit> %d\n", SENDERQUOTALIMIT);
+    logmessage("config: sender timelimit> %d\n", SENDERTIMELIMIT);
+    logmessage("config: sender msgsize> %d\n", SENDERMSGSIZE);
+    logmessage("config: sender expire inactive> %d\n", SENDER_INACTIVE_EXPIRE);
+    logmessage("\n");
+
+    logmessage(" ---- RECIPIENT THROTTLE CONFIG ----\n");
+    logmessage("config: recipient throttle> %d\n", RECIPIENTTHROTTLE);
+    logmessage("config: recipient msglimit> %d\n", RECIPIENTMSGLIMIT);
+    logmessage("config: recipient timelimit> %d\n", RECIPIENTTIMELIMIT);
+    logmessage("config: recipient expire inactive> %d\n", RECIPIENT_INACTIVE_EXPIRE);
+    logmessage("config: postfix_recipient_quota_exceeded> %s\n", postfix_recipient_quota_exceeded);
+    logmessage("config: quota_exceeded_temp_reject> %d\n", QUOTA_EXCEEDED_TEMP_REJECT);
+    logmessage("\n");
   }
 
   if((SENDER_THROTTLE_HOST) && (SENDER_THROTTLE_SASL))
@@ -571,10 +785,38 @@ read_conf(int prog)
     logmessage("FATAL: you may NOT have SENDER_THROTTLE_HOST and SENDER_THROTTLE_SASL enabled\n");
     exit (-1);
   }
+
+  /* quick acl check */
+  if((!CONN_ACL) || (!strlen(CONN_ACL)))
+  {
+    logmessage("FATAL: you did not upgrade correctly or have broken something. "
+               "Please read the Changelog.txt. You're missing the CONN_ACL setting\n");
+    exit (-1);
+  }
 }
 
 
 
+/*
+ * function: w_string_strip
+ *  purpose:
+ *   return:
+ */             
+int
+w_string_strip(void *str, char *token)
+{
+  char	*ptr;
+  ptr	 = strtok(str, token);
+
+  if(ptr != NULL)
+  {
+    return (1); /* found */
+  } else {
+    return (0); /* not found */
+  }
+}
+
+
 
 /* 
  * function: logmessage
@@ -584,7 +826,6 @@ read_conf(int prog)
 void
 logmessage(const char *fmt, ...)
 {
-  
   va_list ap;
   va_start(ap, fmt);
   
@@ -594,7 +835,6 @@ logmessage(const char *fmt, ...)
     vsyslog(SYSLOG_FACILITY, fmt, ap);
   
   va_end(ap);
-  
 }
 
 
@@ -608,11 +848,9 @@ logmessage(const char *fmt, ...)
 void
 usage(char *usag)
 { 
-  
   logmessage("policyd %s\n", VERSION);
   logmessage("usage: %s -c /path/to/policyd.conf\n", usag);
   exit(-1);
-  
 }
 
 
@@ -626,8 +864,7 @@ usage(char *usag)
 int
 extract_seconds(char *token)
 {
-  
-  unsigned char tmp[32];
+           char tmp[32];
   unsigned  int multiplier=0;
 
   /* allow values of 0 */
@@ -672,7 +909,6 @@ extract_seconds(char *token)
   }
   
   strncpy(tmp, token, sizeof(tmp) - 2);
-  
   return (atol(tmp) * multiplier);
 }
 
@@ -685,10 +921,9 @@ extract_seconds(char *token)
  *   return: policy variable (into an array)
  */
 void
-extract(int fd, char *token, int startlen)
+extract(unsigned int fd, char *token, unsigned int startlen)
 {
-  
-  int y, clen=startlen, tlen;
+  unsigned int y, clen=startlen, tlen;
 
   memset(extract_array[fd], 0x00, 64);
   tlen=strlen(token);
@@ -710,22 +945,50 @@ extract(int fd, char *token, int startlen)
       }
     }
   }
-  
 }
 
 
 
 
+/*
+ * function: extract_conf
+ *  purpose: extract policy variable
+ *   return: policy variable (into an array)
+ */
+void
+extract_conf(unsigned int fd, char *token, unsigned int startlen)
+{
+  unsigned int y, clen=startlen, tlen;
+
+  memset(extract_array_conf[fd], 0x00, 512);
+  tlen=strlen(token);
+
+  for( ; clen <= tlen && clen <= 511 ; clen++) {
+    if(token[clen]=='=') {
+      for(clen++,y=0; clen<=tlen&&clen<=511;clen++) {
+        /* we only want characters [A-Z][a-z][0-9]/@ and . */
+        if((isalnum(token[clen]) != 0) || isascii(token[clen] != 0)) {
+	  if((token[clen] != '"') && (token[clen] != '\n')) {
+          extract_array_conf[fd][y]=token[clen]; y++; }
+        }
+      }
+    }
+  }
+}
+
+
+
+
+
 /*
  * function: extract_ip
  *  purpose: extract ip address from policy variable
  *   return: policy variable (into an array)
  */
 void
-extract_ip(int fd, char *token)
+extract_ip(unsigned int fd, char *token)
 {
-  
-  int x=15, y=0, z=0, len;
+  unsigned int x=15, y=0, z=0, len;
 
   memset(extract_ip_array[fd], 0x00, 64);
   len=strlen(token);
@@ -734,15 +997,57 @@ extract_ip(int fd, char *token)
     if(token[x] == '\n') break;
 
     /* we only want characters [0-9] and . */
-    if((isdigit(token[x]) != 0) || (token[x] == '.')) {
-      if(token[x] == '.')
-        z++;
-      if(z == 3)
-        break;
+    if((isdigit(token[x]) != 0) || (token[x] == '.'))
+    {
+      if(token[x] == '.') z++;
+      if(z == GREYLIST_HOSTADDR) break;
       extract_ip_array[fd][y]=token[x]; y++;
     }
   }
+}
+
+
+
+
+/*
+ * function: extract_ip_array
+ *  purpose: extract ip address from policy variable
+ *   return: policy variable (into an array)
+ */
+void
+extract_ipfill(unsigned int fd, char *token)
+{
+  unsigned int x=15, y=0, z=0, len, w=5;
+
+  memset(extract_ip_array[fd], 0x00, 64);
+  len=strlen(token);
+
+  for(x=15,z=0,y=0;x<len||x<64;x++) {
+    if(token[x] == '\n') break;
+
+    /* we only want characters [0-9] and . */
+    if((isdigit(token[x]) != 0) || (token[x] == '.'))
+    {
+      if(token[x] == '.')
+      {
+        if(z == 0) 
+           snprintf(host_array[fd][w], 64, "%s.%%.%%.%%", extract_ip_array[fd]);
+
+        if(z == 1)
+           snprintf(host_array[fd][w], 64, "%s.%%.%%", extract_ip_array[fd]);
   
+        if(z == 2) 
+           snprintf(host_array[fd][w], 64, "%s.%%", extract_ip_array[fd]);
+
+        z++; w--;
+      }
+
+      if(z == GREYLIST_HOSTADDR)
+        break;
+      extract_ip_array[fd][y]=token[x];
+      y++;
+    }
+  }
 }
 
 
@@ -773,7 +1078,6 @@ fold()
 int
 gettime(void)
 {
-
   /* get current time */
   if(gettimeofday(&timevalue, NULL) != 0)
   {
@@ -781,7 +1085,7 @@ gettime(void)
     exit(-1);
   }
 
-  return ((int)timevalue.tv_sec);
+  return (timevalue.tv_sec);
 }
 
 
@@ -795,7 +1099,6 @@ gettime(void)
 void
 drop_privs(void)
 {
-  
   /* 
    * 1) quick sanity check
    * 2) ensure backward compatibility with old configs
@@ -803,12 +1106,12 @@ drop_privs(void)
   if(PIDFILE)
   {
     /* write to pid file */
-    if((pidfile=freopen(PIDFILE, "w", stdout)) == NULL)
+    if((pidfile=fopen(PIDFILE, "w")) == NULL)
     {
       fprintf(stderr, "fopen(): %s: %s\n", strerror(errno), PIDFILE);
       exit(-1);
     }
-    fprintf(pidfile, "%d\n", getpid());
+    fprintf(pidfile, "%d\n", (unsigned int)getpid());
 
     /* we're done, clean up */
     if(fclose(pidfile) != 0)
@@ -819,7 +1122,14 @@ drop_privs(void)
   }
 
   /* change root */
-  if(chroot(CHROOT) == -1)
+  if(chdir(CHROOT) == -1) 
+  {
+    logmessage("chdir(): %s\n", strerror(errno));
+    exit(-1);
+  }
+   
+  /* chroot */
+  if(chroot(".") == -1)
   {
     logmessage("chroot(): %s\n", strerror(errno));
     exit(-1);
@@ -850,8 +1160,10 @@ drop_privs(void)
  *   return: 0=sucessfull write(), 1=failed write()
  */
 int
-policy_reply(int fd, int code)
+policy_reply(unsigned int fd, int code, int status)
 {
+  /* keep gcc quiet for now */
+  status = 0;
   
   /* accept: always allow */
   if(code == 0)
@@ -862,8 +1174,8 @@ policy_reply(int fd, int code)
     {
       
       /* whitelisted */
-        snprintf(xgreylist_array[fd], 128, "%s host: %s whitelisted\n\n",
-	  POSTFIX_X_HEADER, host_array[fd][0]);	
+      snprintf(xgreylist_array[fd], 128, "%s host: %s\n\n",
+        POSTFIX_X_HEADER, host_array[fd][2]);
       
       /* if not whitelisted, greylist.c already filled in all the details */
       w_write(fd, xgreylist_array[fd], strlen(xgreylist_array[fd]));
@@ -902,10 +1214,31 @@ policy_reply(int fd, int code)
       return (0);
   }
   
-  /* reject: max quota exceeded */
+  /* reject: max sender quota exceeded */
   if(code == -5)
   {
-    if(w_write(fd, postfix_quota_exceeded, strlen(postfix_quota_exceeded)) == -1)
+    if(w_write(fd, postfix_sender_quota_exceeded, strlen(postfix_sender_quota_exceeded)) == -1)
+      return (0);
+  }
+
+  /* reject: helo checking */
+  if(code == -6)
+  {
+    if(w_write(fd, postfix_blacklist, strlen(postfix_blacklist)) == -1)
+      return (0);
+  }
+
+  /* reject: max recipient quota exceeded */
+  if(code == -7)
+  {
+    if(w_write(fd, postfix_recipient_quota_exceeded, strlen(postfix_recipient_quota_exceeded)) == -1)
+      return (0);
+  }
+
+  /* reject: max recipient quota exceeded */
+  if(code == -8)
+  {
+    if(w_write(fd, POSTFIX_MODULE_FAILURE, strlen(POSTFIX_MODULE_FAILURE)) == -1)
       return (0);
   }
 
@@ -921,23 +1254,23 @@ policy_reply(int fd, int code)
  *   return: -20 for failures (talk to Postfix first)
  */
 int
-db_failure(int fd, char *module)
+db_failure(unsigned int fd, char *module)
 {
 
   if(FAILSAFE==1)
   {
     /* do not fail */
     logmessage("rcpt=%lu, %s=pass, host=%s (%s), from=%s, to=%s, size=%s\n",
-      rcpt_count,
-      module,
-      host_array[fd][0],        /* host */
-      host_array[fd][1],        /* hostname */
-      triplet_array[fd][1],     /* from */
-      triplet_array[fd][2],     /* rcpt */
-      triplet_array[fd][3]      /* size */
+      rcpt_count,               /* recipient count      */
+      module,                   /* module name          */
+      host_array[fd][2],        /* host                 */
+      host_array[fd][0],        /* hostname             */
+      triplet_array[fd][1],     /* from                 */
+      triplet_array[fd][2],     /* rcpt                 */
+      triplet_array[fd][3]      /* size                 */
     );  
 
-    policy_reply(fd, 0);
+    policy_reply(fd, 0, 0);
     return (-20); 
   }             
                 
@@ -945,16 +1278,16 @@ db_failure(int fd, char *module)
   {             
     /* fail as requested */
     logmessage("rcpt=%lu, %s=fail, host=%s (%s), from=%s, to=%s, size=%s\n",
-      rcpt_count,
-      module,
-      host_array[fd][0],        /* host */
-      host_array[fd][1],        /* hostname */
-      triplet_array[fd][1],     /* from */
-      triplet_array[fd][2],     /* rcpt */
-      triplet_array[fd][3]      /* size */
+      rcpt_count,               /* recipient count      */
+      module,                   /* module name          */
+      host_array[fd][2],        /* host                 */
+      host_array[fd][0],        /* hostname             */
+      triplet_array[fd][1],     /* from                 */
+      triplet_array[fd][2],     /* rcpt                 */
+      triplet_array[fd][3]      /* size                 */
     );
 
-    policy_reply(fd, -1);
+    policy_reply(fd, -1, 0);
     return (-20);
   }
 
