@@ -39,24 +39,54 @@ main(int argc, char **argv)
   char host[48];
   socklen_t clilen;
   struct sockaddr_in cliaddr, servaddr;
+  struct rlimit rlimit_nofile;
 
   if(argc < 2) usage(argv[0]);
-  while ((c = getopt(argc,argv,":c:")) != EOF)
+  while ((c = getopt(argc,argv,":c:v")) != EOF)
   {
     switch(c)
     {
       case 'c':
-	configpath=optarg;
-	read_conf(0);
-	break;
-	
+        configpath=optarg;
+        read_conf(0);
+        break;
+
+      case 'v':
+	logmessage("%s %s\n", PROJECT, VERSION);
+	exit(0);
+
       default:
         usage(argv[0]);
     }
   }
   
   logmessage("starting %s %s\n", PROJECT, VERSION);
+
+  /*
+   *  raise RLIMIT_FSIZE to MAXFDS, bail if cannot.
+   */
+  if (getrlimit(RLIMIT_NOFILE, &rlimit_nofile) == -1)
+  {
+    logmessage("cannot get rlimit: %s\n", strerror(errno));
+    exit (-1);
+  } else {
+    if((MAXFDS+1) > rlimit_nofile.rlim_max)
+    {
+      /* if it needs increasing, keep 1 fd for logging */
+      rlimit_nofile.rlim_max = MAXFDS+1;
+      rlimit_nofile.rlim_cur = MAXFDS+1;
+      if(setrlimit(RLIMIT_NOFILE, &rlimit_nofile) == -1)
+      {
+        logmessage("cannot set rlimit: %s\n", strerror(errno));
+        exit (-1);
+      };
+    };
+  };
   
+  if(DEBUG)
+    logmessage("DEBUG: fd: 0: rlimit: max: %d cur: %d\n",
+      rlimit_nofile.rlim_cur, rlimit_nofile.rlim_max);
+
   /*
    *  bind to port
    */
@@ -70,7 +100,7 @@ main(int argc, char **argv)
   w_listen(listenfd, LISTENQ);
 
   /*
-   *  drop all privledges
+   *  drop all privileges
    */
   drop_privs();
 
@@ -79,17 +109,19 @@ main(int argc, char **argv)
    */
   rcpt_count=0;
   mysql_count=0;
+  mysql_timeout=5;
   mysql_failure_count=0;
   last_mail_time=gettime();
+  if((GREYLIST_HOSTADDR > 4) || (GREYLIST_HOSTADDR < 1))
+    GREYLIST_HOSTADDR=3;
     
   /* 
    *  prefer fd sets
    */
-  maxfd = listenfd;			/* initialize */
-  maxi = -1;				/* index into client[] array */
-
+  maxfd = listenfd;                     /* initialize */
+  maxi = -1;                            /* index into client[] array */
   for(numi=0;numi<MAXFDS;numi++)
-    client[numi] = -1 ;			/* indicates available entry */
+    client[numi] = -1;                  /* indicates available entry */
 
   FD_ZERO(&allset);
   FD_SET(listenfd, &allset);
@@ -108,35 +140,45 @@ main(int argc, char **argv)
   signal(SIGTERM, fold);
   signal(SIGINT, fold);
 
-  /* enter infinate loop */
+  /* enter infinite loop */
   for(;;)
   {
 thunder:
-    rset = allset;                    	/* structure assignment */
+    rset = allset;                      /* structure assignment */
     nready = w_select(maxfd + 1, &rset, NULL,NULL,NULL);
     
-    if(FD_ISSET(listenfd, &rset))    	/* new client connection */
+    if(FD_ISSET(listenfd, &rset))       /* new client connection */
     {
       clilen=sizeof(cliaddr);
       connfd=w_accept(listenfd, (struct sockaddr *) &cliaddr, &clilen);
 
       for(numi=0; numi < MAXFDS; numi++) {
         if(client[numi] < 0) {
-          client[numi] = connfd;       	/* save file descriptor */
-	  if(DEBUG)
-	    logmessage("saved file descriptor: numi = %d, connfd = %d\n", numi, connfd);
+          client[numi] = connfd;        /* save file descriptor */
+          if(DEBUG > 0)
+            logmessage("DEBUG: -> saved file descriptor: numi = %d, connfd = %d\n", numi, connfd);
           break;
         }
       }
 
+      /* tcp acl check */
+      if(w_tcp_conn_acl(w_inet_ntop(AF_INET, &cliaddr.sin_addr, host, sizeof(host))) == -1)
+      {
+	 logmessage("WARNING: connection attempt from: %s\n",
+	   w_inet_ntop(AF_INET, &cliaddr.sin_addr, host, sizeof(host)));
+	 w_close(connfd);
+	 continue;
+      }
+      
+      /* max fds */
       if(numi==MAXFDS)
       {
         logmessage("connection from: %s port: %d slots: %d of %d used (**WARNING**)\n",
           w_inet_ntop(AF_INET, &cliaddr.sin_addr, host, sizeof(host)),
           ntohs(cliaddr.sin_port), numi, MAXFDS);
 
-	w_close(connfd);
-	goto thunder;
+        w_close(connfd);
+        goto thunder;
       }
       
       logmessage("connection from: %s port: %d slots: %d of %d used\n",
@@ -165,12 +207,12 @@ thunder:
         if(DEBUG > 2)
           logmessage("DEBUG: fd: %d select(): activity on fd: %d\n", sockfd, sockfd);
       
-	if(chk_pol(sockfd) == 0)
-	{
+        if(chk_pol(sockfd) == 0)
+        {
           w_close(sockfd);
           FD_CLR(sockfd, &allset);
-          client[numi] = -1;	      /* descriptor available */
-	}
+          client[numi] = -1;          /* descriptor available */
+        }
 
         if(--nready <= 0)
           break;       /* no more readable file descriptors */
@@ -190,7 +232,7 @@ thunder:
  *   return: 0 for connection closing
  */
 int
-chk_pol(int fd)
+chk_pol(unsigned int fd)
 {
   /* clear incase we have stale data */
   bzero(buf[fd], sizeof(buf[fd]));
@@ -220,25 +262,26 @@ chk_pol(int fd)
    *  ensure we have the information needed of enabled modules
    */
   if(module_info_check(fd) == -1)
-    return 0;                      /* fatal: close connection */
-  
+      return(policy_reply(fd, -8, 0));     /* fatal */
+
   /*
-   *  update mail counter, all needed module data is valid
+   *  rcpt: update mail counter, all needed module data is valid
+   *  time: set time to current time
    */
   rcpt_count++;
+  timenow=gettime();
 
   /*
    *  check to see that database connection is live
    */
   if(DATABASE_KEEPALIVE == 1)
   {
-    timenow=gettime();
 
     /*
-     *  do not probe database unless at least 300 seconds 
+     *  do not probe database unless at least 30 seconds 
      *  has passed since the last mail
      */
-    if(timenow >= (last_mail_time+300))
+    if(timenow >= (last_mail_time+30))
     {                             /* timer has expired */
       int probe;
       probe=database_probe(fd);
@@ -249,7 +292,7 @@ chk_pol(int fd)
       if(probe == -20)
       {
         last_mail_time=timenow;
-	goto passthrough;
+        goto passthrough;
       }
     }
 
@@ -262,11 +305,16 @@ chk_pol(int fd)
    * future modules go here: (order matters)
    * 
    *  [X]: whitelisting
+   *  [X]: whitelisting(sender/domain)
+   *  [X]: whitelisting(dnsname)
    *  [X]: blacklisting
-   *  [X]: blacklist helo
+   *  [X]: blacklisting(helo)
+   *  [X]: blacklisting(sender/domain)
    *  [X]: spamtrap
+   *  [X]: helo check (multiple helo forgeries)
    *  [X]: greylisting
    *  [X]: sender throttling
+   *  [X]: recipient throttling
    * 
    */
   
@@ -277,52 +325,117 @@ chk_pol(int fd)
     /* check if whitelisted */
     action_array[fd]=whitelist_check(fd);
     if(action_array[fd] == 1)                 /* whitelisted */
-      return(policy_reply(fd, 0));
+      return(policy_reply(fd, 0, 0));
 
     /* handle database failure */
     if(action_array[fd] == -20)
       return (1);
   }
-	
+
+  /* whitelist */
+  if(WHITELISTSENDER==1)
+  {
+    /* check if sender/domain is whitelisted */
+    action_array[fd]=whitelist_sender_check(fd);
+    if(action_array[fd] == 1)                 /* whitelisted */
+      return(policy_reply(fd, 0, 0));
+
+    /* handle database failure */
+    if(action_array[fd] == -20)
+      return (1);
+  }
+
+  /* whitelist */
+  if(WHITELISTDNSNAME==1)
+  {
+    /* check if sender/domain is whitelisted */
+    action_array[fd]=whitelist_dnsname_check(fd);
+    if(action_array[fd] == 1)                 /* whitelisted */
+      return(policy_reply(fd, 0, 0));
+
+    /* handle database failure */
+    if(action_array[fd] == -20)
+      return (1);
+  }
+
   /* blacklist */
   if(BLACKLISTING==1)
   {
     /* check if blacklisted */
     action_array[fd]=blacklist_check(fd);
     if(action_array[fd] == 1)                 /* blacklisted */
-      return(policy_reply(fd, -2));
+      return(policy_reply(fd, -2, 0));
 
     /* handle database failure */
     if(action_array[fd] == -20)
       return (1);
   }
-	
+
   /* blacklist helo */
   if(BLACKLIST_HELO==1)
   {
     /* check if helo is blacklisted */
     action_array[fd]=blacklist_helo_check(fd);
     if(action_array[fd] == 1)                 /* blacklisted */
-      return(policy_reply(fd, -2));
+      return(policy_reply(fd, -2, 0));
 
     /* handle database failure */
     if(action_array[fd] == -20)
       return (1);
   }
-	  
+  
+  /* blacklist sender/domain */
+  if(BLACKLISTSENDER==1)
+  {
+    /* check if sender/domain is blacklisted */
+    action_array[fd]=blacklist_sender_check(fd);
+    if(action_array[fd] == 1)                 /* blacklisted */
+      return(policy_reply(fd, -2, 0));
+
+    /* handle database failure */
+    if(action_array[fd] == -20)
+      return (1);
+  }
+
+  /* blacklist dnsname */
+  if(BLACKLISTDNSNAME==1)
+  {
+    /* check if sender/domain is blacklisted */
+    action_array[fd]=blacklist_dnsname_check(fd);
+    if(action_array[fd] == 1)                 /* blacklisted */
+      return(policy_reply(fd, -2, 0));
+
+    /* handle database failure */
+    if(action_array[fd] == -20)
+      return (1);
+  }
+
   /* spamtrap */
   if(SPAMTRAPPING==1)
   {
     /* check if spamtrap */
     action_array[fd]=spamtrap_check(fd);
     if(action_array[fd] == 1)                 /* spamtrap */
-      return(policy_reply(fd, -4));
+      return(policy_reply(fd, -4, 0));
 
     /* handle database failure */
     if(action_array[fd] == -20)
       return (1);
   }
 
+  /* helo check */
+  if(HELO_CHECK==1)
+  {
+    /* check if helo is blacklisted */
+    action_array[fd]=helo_check(fd);
+    if(action_array[fd] == 1)                 /* helo abuse */
+      return(policy_reply(fd, -6, 0));
+
+    /* handle database failure */
+    if(action_array[fd] == -20)
+      return (1);
+  }
+  
   /* greylisting */
   if(GREYLISTING==1)
   {
@@ -330,11 +443,11 @@ chk_pol(int fd)
     action_array[fd]=greylist_check(fd);
     
     if(action_array[fd] == 0)                 /* greylist update */
-      if(SENDERTHROTTLE != 1)                 /* dont accept just yet */
-        return(policy_reply(fd, 0));
+      if((SENDERTHROTTLE == 0) && (RECIPIENTTHROTTLE == 0))
+        return(policy_reply(fd, 0, 0));     /* dont accept just yet */
 
     if(action_array[fd] == -1)                /* greylist reject */
-      return(policy_reply(fd, -1));
+      return(policy_reply(fd, -1, 0));
 
     /* handle database failure */
     if(action_array[fd] == -20)
@@ -347,20 +460,38 @@ chk_pol(int fd)
     /* check if sender is throttled */
     action_array[fd]=throttle_check(fd);
 
-    if(action_array[fd] == 0)                 /* allow mail */
-      return(policy_reply(fd, 0));
+    if(action_array[fd] == 0)                     /* allow mail */
+      if(RECIPIENTTHROTTLE == 0)        /* dont accept just yet */
+        return(policy_reply(fd, 0, 0));
     
-    if(action_array[fd] == -3)                /* message size too big */
-      return(policy_reply(fd, -3));
+    if(action_array[fd] == -3)          /* message size too big */
+      return(policy_reply(fd, -3, 0));
 
-    if(action_array[fd] == -5)                /* quota reached */
-      return(policy_reply(fd, -5));
+    if(action_array[fd] == -5)                 /* quota reached */
+      return(policy_reply(fd, -5, 0));
 
     /* handle database failure */
     if(action_array[fd] == -20)
       return (1);
   }
   
+  /* recipient throttling */
+  if(RECIPIENTTHROTTLE==1)
+  {
+    /* check if recipient is throttled */
+    action_array[fd]=throttle_rcpt(fd);
+
+    if(action_array[fd] == 0)                 /* allow mail */
+      return(policy_reply(fd, 0, 0));
+
+    if(action_array[fd] == -7)                /* quota reached */
+      return(policy_reply(fd, -7, 0));
+
+    /* handle database failure */
+    if(action_array[fd] == -20)
+      return (1);
+  }
+      
   /* end of modules */
 
 passthrough:
@@ -370,15 +501,14 @@ passthrough:
    * switch into pass-through mode and allow.
    */
   logmessage("rcpt=%lu, type=passthrough, host=%s (%s), from=%s, to=%s, size=%s\n",
-    rcpt_count,
-    host_array[fd][0],      /* host */
-    host_array[fd][1],      /* hostname */
-    triplet_array[fd][1],   /* from */
-    triplet_array[fd][2],   /* rcpt */
-    triplet_array[fd][3]    /* size */
-  );
+    rcpt_count,                       /* recipient count      */
+    host_array[fd][2],                /* host                 */
+    host_array[fd][0],                /* hostname             */
+    triplet_array[fd][1],             /* from                 */
+    triplet_array[fd][2],             /* rcpt                 */
+    triplet_array[fd][3]);            /* size                 */
   
-  return(policy_reply(fd, 0));
+  return(policy_reply(fd, 0, 0));
 }
 
 
@@ -390,7 +520,7 @@ passthrough:
  *   return: nada
  */
 void
-clear_var(int fd)
+clear_var(unsigned int fd)
 {
   /*
    *  allow by default
@@ -418,7 +548,7 @@ clear_var(int fd)
  *   return: nada
  */
 void
-parse_buf(int fd, char *buf)
+parse_buf(unsigned int fd, char *buf)
 {
 
   /*
@@ -455,8 +585,9 @@ parse_buf(int fd, char *buf)
     if(strncmp(policy_array[fd][i[fd]], "client_address=", 15) == 0)
     {
       extract_ip(fd, policy_array[fd][i[fd]]);
-	            
-      strncpy(host_array[fd][0], policy_array[fd][i[fd]]+15, 64);
+      extract_ipfill(fd, policy_array[fd][i[fd]]);
+
+      strncpy(host_array[fd][2], policy_array[fd][i[fd]]+15, 64);
       strncpy(triplet_array[fd][0], extract_ip_array[fd], 64);
     }
 
@@ -464,33 +595,44 @@ parse_buf(int fd, char *buf)
     if(strncmp(policy_array[fd][i[fd]], "client_name=", 12) == 0)
     {
       extract(fd, policy_array[fd][i[fd]], 11);
-      strncpy(host_array[fd][1], extract_array[fd], 64);
+      strncpy(host_array[fd][0], extract_array[fd], 64);
     }
 
     /* sender address */
     if(strncmp(policy_array[fd][i[fd]], "sender=", 7) == 0)
     {
+      char *s;
       extract(fd, policy_array[fd][i[fd]], 6);
       strncpy(triplet_array[fd][1], extract_array[fd], 64);
 
       /* add in some type of data for the database */
       if(triplet_array[fd][1][0] == 0x00)
-	strncpy(triplet_array[fd][1], "<>", 3); /* append null */
+        strncpy(triplet_array[fd][1], "<>", 3); /* append null */
+
+      /* optinout needed for recipient domain */
+      s=strrchr(extract_array[fd], '@');
+      if(s != NULL)
+      {
+        strncpy(host_array[fd][7], s+1, 64);
+        strncpy(host_array[fd][6], extract_array[fd],
+          strlen(triplet_array[fd][1]) - strlen(host_array[fd][7]) -1);
+      }
     }
 
     /* recipient address */
     if(strncmp(policy_array[fd][i[fd]], "recipient=", 10) == 0)
     {
+      char *r;
       extract(fd, policy_array[fd][i[fd]], 9);
       strncpy(triplet_array[fd][2], extract_array[fd], 60);
                                    /* max length of recipient in mysql */
-      /* optinout on recipient address */
-      if(OPTINOUT == 1)
+      /* optinout needed for recipient domain */
+      r=strrchr(extract_array[fd], '@');
+      if(r != NULL)
       {
-        char *p;
-        p=strrchr(extract_array[fd], '@');
-        if(p != NULL)
-          strncpy(host_array[fd][2], p, 64);
+        strncpy(host_array[fd][9], r+1, 64);
+        strncpy(host_array[fd][8], extract_array[fd],
+          strlen(triplet_array[fd][2]) - strlen(host_array[fd][9]) - 1);
       }
     }
 
@@ -514,6 +656,13 @@ parse_buf(int fd, char *buf)
       extract(fd, policy_array[fd][i[fd]], 9);
       strncpy(triplet_array[fd][5], extract_array[fd], 60);
     }
+    
+    /* instance */
+    if(strncmp(policy_array[fd][i[fd]], "instance=", 9) == 0)
+    {
+      extract(fd, policy_array[fd][i[fd]], 8);
+      strncpy(triplet_array[fd][6], extract_array[fd], 60);
+    }
   }
   
   /* DEBUG: easier when reporting problems */
@@ -538,35 +687,23 @@ parse_buf(int fd, char *buf)
  *   return: 0 for success, -1 for failure
  */
 int
-module_info_check(int fd)
+module_info_check(unsigned int fd)
 {
   
   /* critical information needed for greylisting */
   if(GREYLISTING==1)
   {
     /* ip address */
-    if(host_array[fd][0][0] == 0x00) {
-      logmessage("invalid host_array[fd][0]: (host ip): %s\n", host_array[fd][0]);
+    if(host_array[fd][2][0] == 0x00) {
+      logmessage("invalid host_array[%d][2]: (greylist host ip): %s\n", fd, host_array[fd][2]);
       goto err;
     }
 
     /* recipient address */
     if(triplet_array[fd][2][0] == 0x00) {
-      logmessage("invalid triplet_array[fd][2]: (recipient): %s\n", triplet_array[fd][2]);
+      logmessage("invalid triplet_array[%d][2]: (greylist recipient): %s\n", fd, triplet_array[fd][2]);
       goto err;
     }
-    
-    /* optinout on recipient address */
-    if(OPTINOUT == 1)
-    {
-      char *p;
-      p=strrchr(host_array[fd][2], '@');
-      if(p == NULL)
-      {
-	logmessage("invalid host_array[fd][2]: (recipient): %s\n", host_array[fd][2]);
-	goto err;
-      }
-    }
   }
 
 
@@ -575,30 +712,39 @@ module_info_check(int fd)
   {
     /* sender */
     if(triplet_array[fd][1][0] == 0x00) {
-      logmessage("invalid triplet_array[fd][1]: (sender): %s\n", triplet_array[fd][1]);
+      logmessage("invalid triplet_array[%d][1]: (sender throttle from): %s\n", fd, triplet_array[fd][1]);
       goto err;
     }
     
     /* size */
     if(triplet_array[fd][3][0] == 0x00) {
-      logmessage("invalid triplet_array[fd][3]: (size): %s\n", triplet_array[fd][1]);
+      logmessage("invalid triplet_array[%d][3]: (sender throttle size): %s\n", fd, triplet_array[fd][3]);
       goto err;
     }
   }
 
+  /* critical information needed for recipient throttling */
+  if(RECIPIENTTHROTTLE==1)
+  {
+    /* recipient address */
+    if(triplet_array[fd][2][0] == 0x00) {
+      logmessage("invalid triplet_array[%d][2]: (recipient throttle): %s\n", fd, triplet_array[fd][2]);
+      goto err;
+    }
+  }
   
   /* critical information needed for spamtrapping */
   if(SPAMTRAPPING==1)
   {     
     /* ip address */
-    if(host_array[fd][0][0] == 0x00) {
-      logmessage("invalid host_array[fd][0]: (host ip): %s\n", host_array[fd][0]);
+    if(host_array[fd][2][0] == 0x00) {
+      logmessage("invalid host_array[%d][2]: (spamtrap host ip): %s\n", fd, host_array[fd][2]);
       goto err;
     }
       
     /* recipient address */        
     if(triplet_array[fd][2][0] == 0x00) {
-      logmessage("invalid triplet_array[fd][2]: (recipient): %s\n", triplet_array[fd][2]);
+      logmessage("invalid triplet_array[%d][2]: (spamtrap recipient): %s\n", fd, triplet_array[fd][2]);
       goto err;
     }
   }  
@@ -606,9 +752,9 @@ module_info_check(int fd)
   /* critical information needed for helo blacklisting */
   if(BLACKLIST_HELO==1)
   {     
-    /* recipient address */        
+    /* helo information */        
     if(triplet_array[fd][5][0] == 0x00) {
-      logmessage("invalid triplet_array[fd][2]: (helo): %s\n", triplet_array[fd][2]);
+      logmessage("invalid triplet_array[%d][5]: (blacklist helo): %s\n", fd, triplet_array[fd][5]);
       goto err;
     }
   }  
